// SPDX-License-Identifier: AGPL-3.0-or-later
// SPDX-FileCopyrightText: 2022-2023 grommunio GmbH
// This file is part of Gromox.
#define _GNU_SOURCE 1 /* strcasestr */
#include <atomic>
#include <cerrno>
#include <cstdio>
#include <cstring>
#include <memory>
#include <tinyxml2.h>
#include <utility>
#include <unordered_map>
#include <gromox/fileio.h>
#include <gromox/paths.h>
#include <gromox/util.hpp>
#include <libHX/io.h>
#include <libHX/string.h>

using namespace gromox;

static std::atomic<unsigned int> wintz_loaded;
static std::unordered_map<std::string, std::string> iana_to_wzone;
static std::unordered_map<std::string, std::string> wzone_to_tzdef;

static errno_t wintz_load_namemap(const char *dirs)
{
	std::unique_ptr<FILE, file_deleter> fp(fopen_sd("windowsZones.xml", dirs));
	if (fp == nullptr) {
		mlog(LV_ERR, "Could not open windowsZones.xml: %s", strerror(errno));
		return errno;
	}
	tinyxml2::XMLDocument doc;
	auto ret = doc.LoadFile(fp.get());
	if (ret != tinyxml2::XML_SUCCESS) {
		mlog(LV_ERR, "Failed to load/parse windowsZones.xml");
		return EIO;
	}
	auto node = doc.RootElement();
	if (node == nullptr)
		return EIO;
	auto name = node->Name();
	if (name == nullptr || strcasecmp(name, "supplementalData") != 0) {
		mlog(LV_ERR, "No supplemental root element");
		return EIO;
	}
	node = node->FirstChildElement("windowsZones");
	if (node == nullptr) {
		mlog(LV_ERR, "No windowsZones element");
		return EIO;
	}
	node = node->FirstChildElement("mapTimezones");
	if (node == nullptr) {
		mlog(LV_ERR, "No mapTimezones element");
		return EIO;
	}
	for (node = node->FirstChildElement("mapZone");
	     node != nullptr; node = node->NextSiblingElement("mapZone")) {
		auto oattr = node->FindAttribute("other");
		auto iattr  = node->FindAttribute("type");
		if (oattr == nullptr || iattr == nullptr)
			continue;
		auto oval = oattr->Value();
		auto ival = iattr->Value();
		if (oval == nullptr || ival == nullptr)
			continue;
		std::string ovs = oval, ivs = ival;
		HX_strlower(ivs.data());
		auto p = strcasestr(ovs.c_str(), " Standard Time");
		if (p != nullptr)
			ovs.erase(p - ovs.c_str(), 14);
		replace_unsafe_basename(ovs.data());
		for (auto &&i : gx_split(ivs, ' '))
			if (!i.empty())
				iana_to_wzone.emplace(std::move(i), ovs);
	}
	return 0;
}

static errno_t wintz_load_tzdefs(const char *dirs)
{
	for (const auto &tzpair : iana_to_wzone) {
		auto &wzone = tzpair.second;
		std::unique_ptr<FILE, file_deleter> fp(fopen_sd((wzone + ".tzd").c_str(), dirs));
		if (fp == nullptr) {
			mlog(LV_ERR, "Could not open %s: %s",
			        wzone.c_str(), strerror(errno));
			return errno;
		}
		size_t sl = 0;
		std::unique_ptr<char[], stdlib_delete> sd(HX_slurp_fd(fileno(fp.get()), &sl));
		if (sd == nullptr) {
			mlog(LV_ERR, "slurp_fd: %s", strerror(errno));
			return errno;
		}
		wzone_to_tzdef.emplace(wzone, std::string(sd.get(), sl));
	}
	return 0;
}

static errno_t wintz_load_once(const char *dirs)
{
	unsigned int exp = 0;
	if (!wintz_loaded.compare_exchange_strong(exp, 1))
		return 0;
	auto ret = wintz_load_namemap(dirs);
	if (ret != 0)
		return ret;
	ret = wintz_load_tzdefs(dirs);
	if (ret != 0)
		return ret;
	return 0;
}

namespace gromox {

const std::string *wintz_to_tzdef(const char *izone, const char *dirs)
{
	if (dirs == nullptr)
		dirs = PKGDATADIR;
	if (wintz_load_once(dirs) != 0)
		return nullptr;
	auto ti = wzone_to_tzdef.find(izone);
	if (ti == wzone_to_tzdef.end())
		return nullptr;
	return &ti->second;
}

const std::string *ianatz_to_tzdef(const char *izone, const char *dirs)
{
	if (dirs == nullptr)
		dirs = PKGDATADIR;
	if (wintz_load_once(dirs) != 0)
		return nullptr;
	std::string lizone = izone;
	HX_strlower(lizone.data());
	auto wi = iana_to_wzone.find(lizone);
	if (wi == iana_to_wzone.end())
		return nullptr;
	return wintz_to_tzdef(wi->second.c_str());
}

}
